@@ -108,6 +108,14 @@ def main(argv=None):
     config_parser.add_argument('-c', '--create', action='store_true', default=False,
                                dest='create_config',
                                help='if no local configuration is found, then create one')
+    config_parser.add_argument('--add-terminate-prog', metavar='PATH',
+                               dest='add_terminate_prog', default=None,
+                               help='add program to list of commands to execute')
+    config_parser.add_argument('--rm-terminate-prog', metavar='PATH',
+                               dest='rm_terminate_prog', default=None,
+                               help=('remove program from list of commands to execute; '
+                                     'for example, '
+                                     'copy-and-paste value shown in `hardshare config -l` here'))
     config_parser.add_argument('--add-key', metavar='FILE',
                                dest='new_api_token',
                                help='add new account key')
@@ -123,6 +131,12 @@ def main(argv=None):
     config_parser.add_argument('--rm-raw-device', metavar='PATH', type=str,
                                dest='remove_raw_device_path', default=None,
                                help='remove device previously marked for inclusion in container')
+    config_parser.add_argument('--add-init-inside', metavar='CMD', type=str,
+                               dest='add_init_inside', default=None,
+                               help='add command to be executed inside container')
+    config_parser.add_argument('--rm-init-inside', action='store_true', default=False,
+                               dest='rm_init_inside',
+                               help='remove (empty) list of commands for inside initialization')
     config_parser.add_argument('-p', '--prune', action='store_true', default=False,
                                dest='prune_err_keys',
                                help=('delete files in local key directory that'
@@ -290,6 +304,8 @@ def main(argv=None):
             os.close(0)
             os.close(1)
             os.close(2)
+        else:
+            logger.addHandler(logging.StreamHandler())
         return ac.run_sync(config['wdeployments'][index]['id'])
 
     elif argv_parsed.command == 'terminate':
@@ -319,7 +335,11 @@ def main(argv=None):
             if ac is None:
                 print('cannot terminate without valid API client')
                 return 1
-            ac.terminate(config['wdeployments'][index]['id'])
+            try:
+                ac.terminate(config['wdeployments'][index]['id'])
+            except FileNotFoundError:
+                print('ERROR: cannot reach daemon. Does it exist? (Try `hardshare status`)')
+                return 1
             return 0
 
     elif argv_parsed.command == 'register':
@@ -425,6 +445,10 @@ def main(argv=None):
                             wdeployment['cargs'],
                             wdeployment['image'],
                         ))
+                        if wdeployment['terminate']:
+                            print('\tterminate:')
+                            for terminate_p in wdeployment['terminate']:
+                                print('\t\t{}'.format(terminate_p))
 
                 print('\nfound keys:')
                 if len(config['local']['keys']) == 0:
@@ -509,6 +533,30 @@ def main(argv=None):
             config['wdeployments'][index]['cargs'].remove(carg)
             modify_local(config)
 
+        elif argv_parsed.add_init_inside is not None:
+            config, index, rc = get_config_with_index(argv_parsed.id_prefix)
+            if rc != 0:
+                return rc
+            cprovider = config['wdeployments'][index]['cprovider']
+            if cprovider not in ['docker', 'podman']:
+                print('unknown cprovider: {}'.format(cprovider))
+                return 1
+
+            config['wdeployments'][index]['init_inside'].append(argv_parsed.add_init_inside)
+            modify_local(config)
+
+        elif argv_parsed.rm_init_inside:
+            config, index, rc = get_config_with_index(argv_parsed.id_prefix)
+            if rc != 0:
+                return rc
+            cprovider = config['wdeployments'][index]['cprovider']
+            if cprovider not in ['docker', 'podman']:
+                print('unknown cprovider: {}'.format(cprovider))
+                return 1
+
+            config['wdeployments'][index]['init_inside'] = []
+            modify_local(config)
+
         elif argv_parsed.cprovider_img is not None:
             config, index, rc = get_config_with_index(argv_parsed.id_prefix)
             if rc != 0:
@@ -517,13 +565,45 @@ def main(argv=None):
             if cprovider not in ['docker', 'podman']:
                 print('unknown cprovider: {}'.format(cprovider))
                 return 1
-            cp_images = subprocess.run([cprovider, 'image', 'exists', argv_parsed.cprovider_img])
-            if cp_images.returncode != 0:
-                print('ERROR: given image name is not recognized by cprovider')
-                return 1
+
+            if cprovider == 'podman':
+                cp_images = subprocess.run([cprovider, 'image', 'exists', argv_parsed.cprovider_img])
+                if cp_images.returncode != 0:
+                    print('ERROR: given image name is not recognized by cprovider')
+                    return 1
+            else:  # cprovider == 'docker'
+                cp_images = subprocess.run([cprovider, 'image', 'inspect', argv_parsed.cprovider_img],
+                                           stdout=subprocess.DEVNULL,
+                                           stderr=subprocess.DEVNULL)
+                if cp_images.returncode != 0:
+                    print('ERROR: given image name is not recognized by cprovider')
+                    return 1
+
             config['wdeployments'][index]['image'] = argv_parsed.cprovider_img
             modify_local(config)
 
+        elif argv_parsed.add_terminate_prog is not None:
+            config, index, rc = get_config_with_index(argv_parsed.id_prefix)
+            if rc != 0:
+                return rc
+
+            normalized_path = os.path.abspath(argv_parsed.add_terminate_prog)
+
+            if not os.path.exists(normalized_path):
+                print('ERROR: given path does not exist')
+                return 1
+
+            config['wdeployments'][index]['terminate'].append(normalized_path)
+            modify_local(config)
+
+        elif argv_parsed.rm_terminate_prog is not None:
+            config, index, rc = get_config_with_index(argv_parsed.id_prefix)
+            if rc != 0:
+                return rc
+
+            config['wdeployments'][index]['terminate'].remove(argv_parsed.rm_terminate_prog)
+            modify_local(config)
+
         else:
             print('Use `hardshare config` with a switch. For example, `hardshare config -l`')
             print('or to get a help message, enter\n\n    hardshare help config')
