@@ -31,7 +31,7 @@ logger = logging.getLogger(__name__)
 
 
 class WorkspaceInstance:
-    def __init__(self, cprovider=None, cargs=None, image=None, event_loop=None):
+    def __init__(self, cprovider=None, cargs=None, image=None, terminate=None, event_loop=None):
         if event_loop is None:
             self.loop = asyncio.get_event_loop()
         else:
@@ -45,6 +45,7 @@ class WorkspaceInstance:
             self.img = 'rerobots/hs-generic'
         else:
             self.img = image
+        self.terminate = terminate
         self.status = 'INIT'
         self.container_name = 'rrc' + str(random.randint(0, 100000))  # TODO: check for existing container with this name
         self.instance_id = None
@@ -86,6 +87,10 @@ class WorkspaceInstance:
                 container_name = str(msg[7:-1], encoding='utf-8')
             else:
                 raise ValueError('Unknown daemon status: {}'.format(msg))
+        except ConnectionRefusedError:
+            if 'warnings' not in findings:
+                findings['warnings'] = []
+            findings['warnings'].append('fail to open daemon socket at {}'.format(to_addr))
         except socket.timeout:
             pass
         except BrokenPipeError:
@@ -119,15 +124,21 @@ class WorkspaceInstance:
                             findings['container']['image_tags'] = iminfo['RepoTags']
                     else:
                         findings['has_instance'] = False
-                        findings['warnings'] = 'container name {} not found with cprovider {}'.format(container_name, cprovider)
+                        if 'warnings' not in findings:
+                            findings['warnings'] = []
+                        findings['warnings'].append('container name {} not found with cprovider {}'.format(container_name, cprovider))
                 except FileNotFoundError:
                     findings['has_instance'] = False
-                    findings['warnings'] = 'cprovider {} not found. Is it installed?'.format(cprovider)
+                    if 'warnings' not in findings:
+                        findings['warnings'] = []
+                    findings['warnings'].append('cprovider {} not found. Is it installed?'.format(cprovider))
             else:
                 findings['has_instance'] = False
         else:
             findings['has_instance'] = False
-            findings['warnings'] = 'cprovider "{}" not known'.format(cprovider)
+            if 'warnings' not in findings:
+                findings['warnings'] = []
+            findings['warnings'].append('cprovider "{}" not known'.format(cprovider))
         return findings
 
 
@@ -272,7 +283,8 @@ class WorkspaceInstance:
         await ws_send(json.dumps({
             'v': 0,
             'cmd': 'INSTANCE_STATUS',
-            's': self.status
+            's': self.status,  # == READY
+            'h': self.hostkey,
         }))
 
         try:
@@ -294,10 +306,14 @@ class WorkspaceInstance:
                     logger.warning('tunnel subprocess failed '
                                    'with nonzero exit code: {}'
                                    .format(sshtunnel.returncode))
-            await ws_send(json.dumps({
-                'v': 0,
-                'cmd': 'SSHTUN_DELETE',
-            }))
+            try:
+                await ws_send(json.dumps({
+                    'v': 0,
+                    'cmd': 'SSHTUN_DELETE',
+                }))
+            except Exception as e:
+                logger.warning('caught exception while trying to send SSHTUN_DELETE via WebSocket; '
+                               '{}: {}'.format(type(e), e))
 
 
     async def start_vpn(self, ws_send, ws_recv):
@@ -377,7 +393,8 @@ class WorkspaceInstance:
             await ws_send(json.dumps({
                 'v': 0,
                 'cmd': 'INSTANCE_STATUS',
-                's': self.status
+                's': self.status,  # == READY
+                'h': self.hostkey,
             }))
 
         except:
@@ -406,10 +423,13 @@ class WorkspaceInstance:
             }))
 
 
-    async def launch_instance(self, instance_id, ws_send, ws_recv, conntype, initial_publickey, tunnelkey_path=None):
+    async def launch_instance(self, instance_id, ws_send, ws_recv, conntype, initial_publickey, init_inside=None, tunnelkey_path=None):
         if self.cprovider not in ['docker', 'podman']:
             raise ValueError('unknown cprovider: {}'.format(self.cprovider))
 
+        if init_inside is None:
+            init_inside = []
+
         self.conntype = conntype
         self.instance_id = instance_id
 
@@ -437,9 +457,12 @@ class WorkspaceInstance:
                 launch_args.extend(['-p', '127.0.0.1::22'])
             launch_args += [self.img]
             logger.debug('subprocess: {}'.format(launch_args))
-            subprocess.check_call(launch_args,
-                                  stdout=subprocess.DEVNULL,
-                                  stderr=subprocess.DEVNULL)
+            p = await asyncio.create_subprocess_exec(*launch_args,
+                                                     stdout=subprocess.DEVNULL,
+                                                     stderr=subprocess.DEVNULL)
+            rc = await p.wait()
+            if rc != 0:
+                raise Exception('command ({}) returen error code: {}'.format(launch_args, rc))
 
             if self.cprovider == 'docker':
                 self.container_addr = await self.get_container_addr(timeout=10)
@@ -448,7 +471,6 @@ class WorkspaceInstance:
                 self.container_addr = '127.0.0.1'
                 self.container_port_ssh = await self.get_container_sshport(timeout=10)
 
-            self.hostkey = await self.get_container_hostkey(timeout=45)
             assert self.container_addr is not None
 
             cexec = [self.cprovider, 'exec', self.container_name]
@@ -460,12 +482,26 @@ class WorkspaceInstance:
                                 cexec + ['/bin/chown', '0:0', '/root/.ssh/authorized_keys']]
             for command in prepare_commands + movekey_commands:
                 logger.debug('subprocess: {}'.format(command))
-                subprocess.check_call(command,
-                                      stdout=subprocess.DEVNULL,
-                                      stderr=subprocess.DEVNULL)
+                p = await asyncio.create_subprocess_exec(*command,
+                                                         stdout=subprocess.DEVNULL,
+                                                         stderr=subprocess.DEVNULL)
+                rc = await p.wait()
+                if rc != 0:
+                    raise Exception('command ({}) returned error code: {}'.format(command, rc))
 
             os.unlink(fname)
 
+            self.hostkey = await self.get_container_hostkey(timeout=45)
+
+            for command in init_inside:
+                logger.debug('init inside: {}'.format(command))
+                p = await asyncio.create_subprocess_exec(*(cexec + ['/bin/bash', '-c', command]),
+                                                         stdout=subprocess.DEVNULL,
+                                                         stderr=subprocess.DEVNULL)
+                rc = await p.wait()
+                if rc != 0:
+                    raise Exception('command ({}) returned error code: {}'.format(command, rc))
+
         except Exception as e:
             logger.error('caught exception {}: {}'.format(type(e), e))
             self.status = 'INIT_FAIL'
@@ -475,6 +511,18 @@ class WorkspaceInstance:
             'cmd': 'INSTANCE_STATUS',
             's': self.status
         }))
+        await self.start_tunnel_task(ws_send, ws_recv)
+
+
+    async def start_tunnel_task(self, ws_send, ws_recv, stop_existing=False):
+        if self.tunnel_task is not None:
+            if not stop_existing:
+                raise ValueError('start_tunnel_task called with existing tunnel task, '
+                                 'but stop_existing=False')
+            self.tunnel_task.cancel()
+            while not self.tunnel_task.done():
+                await asyncio.sleep(0.5)
+            self.tunnel_task = None
         if self.conntype == 'vpn':
             self.tunnel_task = self.loop.create_task(self.start_vpn(ws_send, ws_recv))
         else:  # self.conntype == 'sshtun'
@@ -483,11 +531,29 @@ class WorkspaceInstance:
 
     async def destroy_instance(self):
         destroy_args = [self.cprovider, 'rm', '-f', self.container_name]
-        subprocess.check_call(destroy_args,
-                              stdout=subprocess.DEVNULL,
-                              stderr=subprocess.DEVNULL)
+        destroy_p = await asyncio.create_subprocess_exec(*destroy_args,
+                                                         stdout=subprocess.DEVNULL,
+                                                         stderr=subprocess.DEVNULL)
+        rc = await destroy_p.wait()
+        if rc != 0:
+            raise Exception('command ({}) returned error code: {}'.format(destroy_args, rc))
         if self.tunnel_task is not None:
             self.tunnel_task.cancel()
             while not self.tunnel_task.done():
                 await asyncio.sleep(0.5)
             self.tunnel_task = None
+        if self.terminate is not None:
+            for terminate_command in self.terminate:
+                logger.info('starting termination command: {}'.format(terminate_command))
+                try:
+                    tp = await asyncio.create_subprocess_exec(terminate_command,
+                                                              stdout=subprocess.DEVNULL,
+                                                              stderr=subprocess.DEVNULL)
+                except FileNotFoundError:
+                    logger.error('file not found for termination command ({}); skipping...'.format(terminate_command))
+                    continue
+                rc = await tp.wait()
+                if rc != 0:
+                    logger.error('termination command ({}) failed with exitcode {}'.format(terminate_command, rc))
+                else:
+                    logger.info('termination command ({}) completed successfully'.format(terminate_command))
